/** -*- c++ -*- **************************************************************
 **
 ** Copyright 2014 Daniel Paschka (daniel@drachenschuppe.de)
 **
 ** Licensed under the Apache License, Version 2.0 (the "License");
 ** you may not use this file except in compliance with the License.
 ** You may obtain a copy of the License at
 **
 ** http://www.apache.org/licenses/LICENSE-2.0
 **
 ** Unless required by applicable law or agreed to in writing, software
 ** distributed under the License is distributed on an "AS IS" BASIS,
 ** WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 ** See the License for the specific language governing permissions and
 ** limitations under the License.
 *****************************************************************************/

#define __STDC_FORMAT_MACROS
#include <inttypes.h>
#include <stdint.h>
#include <iostream>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>

#include "gdbbackend.h"
#include "targetinterface.h"

using namespace std;

gdbbackend::gdbbackend(std::iostream *ios, targetif *tgt)
{
	this->tgt = tgt;
	this->ios = ios;
}

void gdbbackend::send(const char *fmt, ...)
{
	char buf[4096];
	char cksumStr[3] = {0};

	va_list v;
	va_start(v, fmt);
	int len = vsnprintf(buf, sizeof(buf)-1, fmt, v);
	va_end(v);

	// append checksum
	unsigned char cksum = 0;
	int i;
	for (i=0; i<len; i++)
		cksum += buf[i];
	sprintf(cksumStr, "%02x", cksum);

	// generate ouput in the form of: $str#chksum
	*ios << '$' << buf << '#' << cksumStr;
}

int gdbbackend::dollarCmd(char *cmd, size_t len)
{
	// cout << cmd << "\n";

	/*********************************************************
	 *                                                       *
	 *                    query command                      *
	 *                                                       *
	 *********************************************************/

	if (!strncmp(cmd, "qPart:sysreg:read::", 19))
	{
		char *endptr;
		int reg = strtoll(cmd + 19, &endptr, 16);
		// int len = strtoll(endptr + 1, &endptr, 16);

		uint32_t val = 0;
		tgt->read_system_register(reg, val);
		send("%08"PRIx32, val);

		return 0;
	}

	if (!strncmp(cmd, "qPart:sysreg:write::", 20))
	{
		char *endptr;
		int      reg = strtoll(cmd + 20, &endptr, 16);
		uint32_t val = strtoll(endptr + 1, &endptr, 16);

		tgt->write_system_register(reg, val);
		tgt->read_system_register(reg, val);
		send("%08"PRIx32, val);

		return 0;
	}

	if (!strcmp(cmd, "qSupported"))
	{
		send("qXfer:memory-map:read+;PacketSize=256");
		return 0;
	}

	if (!strcmp(cmd, "vCont?"))
	{
		send("vCont;c;s");
		return 0;
	}

	// attachted to existing process?
	if (!strcmp(cmd, "qAttached"))
	{
		send("0");
		return 0;
	}

	if (!strncmp(cmd, "qXfer:memory-map:read::", 23))
	{
		char *endptr;
		int off    = strtoll(cmd + 24, &endptr, 16);
		int reqlen = strtoll(endptr + 1, &endptr, 16);
		int len    = strlen(targetif::memory_map);

		char buf[4096];
		buf[0] = (len > off + reqlen) ? 'm' : 'l';
		memcpy(buf + 1, targetif::memory_map + off, reqlen);
		send(buf);

		return 0;
	}

	if (cmd[0] == '!')
	{
		send("OK");
		return 0;
	}

	/*********************************************************
	 *                                                       *
	 *                    flash handling                     *
	 *                                                       *
	 *********************************************************/

	if (!strncmp(cmd, "vFlashErase", 12))
	{
		char    *endptr = cmd + 12;
		uint32_t addr   = strtoll(endptr + 1, &endptr, 16);
		int      len    = strtoll(endptr + 1, &endptr, 16);
		int      ret;

		ret = tgt->flashErase(addr, len);
		send((ret >= 0) ? "OK" : "");
		return 0;
	}

	if (!strncmp(cmd, "vFlashWrite", 12))
	{
		char    *endptr = cmd + 12;
		uint32_t addr   = strtoll(endptr + 1, &endptr, 16);
		unsigned i, j;
		int      ret;

		endptr++;
		for (i=0, j=0; j<len - (endptr-cmd);)
		{
			unsigned char c = endptr[j++];
			if (c != 0x7d)
				endptr[i++] = c;
			else
				endptr[i++] = endptr[j++]^0x20;
		}

		ret = tgt->flashWrite(addr, i, (unsigned char*)endptr);
		send((ret >= 0) ? "OK" : "");
		return 0;
	}

	if (!strncmp(cmd, "vFlashDone", 12))
	{
		tgt->flashFinish();
		send("OK");
		return 0;
	}

	/*********************************************************
	 *                                                       *
	 *                running stopping, etc.                 *
	 *                                                       *
	 *********************************************************/

	// reason for target halt
	if (cmd[0] == '?')
	{
		send("S%02x", tgt->last_stop_sig);
		return 0;
	}

	if (cmd[0] == 'c' || !strcmp(cmd, "vCont;c"))
	{
		tgt->cont();
		return 0;
	}

	if (cmd[0] == 's' || !strcmp(cmd, "vCont;s"))
	{
		tgt->step();
		return 0;
	}

	// kill
	if ((cmd[0] == 'k') || (cmd[1] == 'R'))
	{
		tgt->reset();
		return 0;
	}

	/*********************************************************
	 *                                                       *
	 *                break and watchpoints                  *
	 *                                                       *
	 *********************************************************/

	// insert/remove breakpoint
	if ((cmd[0] == 'Z') || (cmd[0] == 'z'))
	{
		char    *endptr = cmd;
		int      type   = strtoll(endptr + 1, &endptr, 16);
		uint32_t addr   = strtoll(endptr + 1, &endptr, 16);
//		int      align  = strtoll(endptr + 1, &endptr, 16);

		int ret = 0;
		if (cmd[0] == 'Z') ret = tgt->breakpoint_set   (type, addr);
		if (cmd[0] == 'z') ret = tgt->breakpoint_remove(type, addr);
		send((ret >= 0) ? "OK" : "");

		return 0;
	}

	/*********************************************************
	 *                                                       *
	 *                replies to data inspect                *
	 *                                                       *
	 *********************************************************/

	// G

	// read/write memory
	if ((cmd[0] == 'm') || (cmd[0] == 'M'))
	{
		char    *endptr = cmd;
		uint32_t addr   = strtoll(endptr + 1, &endptr, 16);
		int      len    = strtoll(endptr + 1, &endptr, 16);

		unsigned char data[4096];
		int i;
		int ret = 0;

		if (cmd[0] == 'm')
		{
			ret = tgt->memory_read(addr, len, data);
			if (ret < 0)
				send("");
			else
			{
				char buf[4096];
				for (i=0; i<len; i++)
					sprintf(buf + i*2, "%02x", data[i]);
				buf[i*2] = 0;
				send(buf);
			}
		}

		if (cmd[0] == 'M')
		{
			char value[3] = { 0 };

			endptr++;
			for (i=0; i<len; i++)
			{
				value[0] = *endptr++;
				if (!value[0]) break;
				value[1] = *endptr++;
				if (!value[1]) break;

				data[i] = strtoll(value, NULL, 16);
			}

			ret = tgt->memory_write(addr, len, data);
			send((ret >= 0) ? "OK" : "");
		}

		return 0;
	}

	// read single register
	if (cmd[0] == 'p')
	{
		char *endptr = cmd;
		int reg = strtol(endptr + 1, &endptr, 16);

		uint32_t val = 0;
		tgt->read_mcu_register(reg, val);
		send("%08"PRIx32, val);
		return 0;
	}

	// write single register
	if (cmd[0] == 'P')
	{
		char    *endptr = cmd;
		int      reg = strtoll(endptr + 1, &endptr, 16);
		uint32_t val = strtoll(endptr + 1, &endptr, 16);

		int ret = tgt->write_mcu_register(reg, val);
		send((ret >= 0) ? "OK" : "");
		return 0;
	}

	// read all registers
	if (cmd[0] == 'g')
	{
		int i;
		char buf[8 * 16 + 1] = { 0 };

		for (i=0; i<16; i++)
		{
			uint32_t val = 0;
			tgt->read_mcu_register(i, val);
			sprintf(buf + i*8, "%08"PRIx32, val);
		}

		send(buf);
		return 0;
	}

	return -1;
}

void gdbbackend::run()
{
	while (!ios->eof())
		if (ios->rdbuf()->in_avail())
		{
			char c = ios->get();
			switch (c)
			{
				case  -1:
					return;

				case '+':
				case '-':
//					printf("got ACK/NACK (%c)\n", c);
				break;

				case   3: // break
//					cout << "break signaled!\n";
					tgt->stop();
				break;

				case '$':
				{
					char buf[1024];
					ios->getline(buf, sizeof(buf), '#');

					// read the checksum
					char csum[3] = {0};
					int  csum_recv;
					csum[0] = ios->get();
					csum[1] = ios->get();
					sscanf(csum, "%02x", &csum_recv);

					// calculate the checksum of the data
					uint8_t csum_calc = 0;
					int i;
					for (i=0; buf[i]; i++)
						csum_calc += buf[i];

//					printf("%s => chksum: 0x%02x == 0x%02x - ", buf, csum_recv, csum_calc );
					if (csum_recv == csum_calc)
					{
						*ios << "+"; // ACK
						if (dollarCmd(buf, i))
						{
							cout << "command \"" << buf << "\" not understood" << "\n";
							send("");
						}
					}
					else
						*ios << "-"; // NACK
				}
				break;

				default:
					cout << "What is " << (char)c << " ?\n";
			} // switch
		}
		else
			if (tgt->check_stop())
				send("S%02x");
} // run

