/** -*- c++ -*- **************************************************************
 **
 ** Copyright 2014 Daniel Paschka (daniel@drachenschuppe.de)
 **
 ** Licensed under the Apache License, Version 2.0 (the "License");
 ** you may not use this file except in compliance with the License.
 ** You may obtain a copy of the License at
 **
 ** http://www.apache.org/licenses/LICENSE-2.0
 **
 ** Unless required by applicable law or agreed to in writing, software
 ** distributed under the License is distributed on an "AS IS" BASIS,
 ** WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 ** See the License for the specific language governing permissions and
 ** limitations under the License.
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>

#include "tcpstream.h"
#include "gdbbackend.h"
#include "targetinterface.h"
#include "../jtag/jtagAdapter.h"
#include "../jtag/Device_AVR32.h"

#include "../jtag/AdapterXilinxPar.h"


struct args_st
{
	unsigned help           :1;
	unsigned version        :1;
	unsigned reset_tgt      :1;
	unsigned chiperase_tgt  :1;
	unsigned transaction    :1;
	unsigned gui            :1;
	unsigned debug_level    :5;

	char *filename;
	char *comport;
	char *emulator;
	char *sim_part;
	char *part;
	char *flash;
	unsigned short gdbaddress;
	unsigned short traceaddress;
	struct {
		int devbefore;
		int devafter;
		int bitsbefore;
		int bitsafter;
	} chain;
} arguments;

int parse_argv(int argc, char **argv)
{
	int c;
	memset (&arguments, 0, sizeof(arguments));

	while (1)
	{
		static struct option long_options[] =
		{
			{"help",         no_argument,       0, 'h'},
			{"version",      no_argument,       0, 'V'},
			{"no-reset",     no_argument,       0, 'R'},
			{"chiperase",    no_argument,       0, 'E'},
			{"transactions", no_argument,       0, 'T'},
			{"debug",        optional_argument, 0, 'd'},
			{"traceaddress", required_argument, 0, 't'},
			{"gdbaddress",   required_argument, 0, 'a'},
			{"simulate",     required_argument, 0, 's'},
			{"file",         required_argument, 0, 'F'},
			{"comport",      required_argument, 0, 'c'},
			{"emulator",     required_argument, 0, 'e'},
			{"daisy",        required_argument, 0, 'D'},
			{"gui",          required_argument, 0, 'g'},
			{"part",         required_argument, 0, 1  },
			{"clock",        required_argument, 0, 2  },
			{0, 0, 0, 0}
		};
		/* getopt_long stores the option index here. */
		int option_index = -1;

		c = getopt_long (argc, argv, "hVRETd::t:a:s:F:c:e:D:f:", long_options, &option_index);

		/* Detect the end of the options. */
		if (c == -1)
			break;

		switch (c)
		{
			case 'h' : arguments.help          = 1; break;
			case 'V' : arguments.version       = 1; break;
			case 'R' : arguments.reset_tgt     = 1; break;
			case 'E' : arguments.chiperase_tgt = 1; break;
			case 'T' : arguments.transaction   = 1; break;
			case 'g' : arguments.gui           = 1; break;

			case 'd' :
				arguments.debug_level = 1;
				if (optarg)
				{
					char *p, l = strlen(optarg);
					arguments.debug_level = strtol(optarg, &p, 0);

					if (p - optarg != l)
					{
						printf("This \"%s\" is not a number!\n", optarg);
						return -2;
					}
				}
			break;

			case 't' :
			{
				char *p = strchr(optarg , ':');
				if (!p) p = optarg;
				else p++;

				arguments.traceaddress = atoi(p);
			}
			break;

			case 'a' :
			{
				char *p = strchr(optarg , ':');
				if (!p) p = optarg;
				else p++;

				arguments.gdbaddress = atoi(p);
			}
			break;

			case 's' :
				arguments.sim_part = optarg;
			break;

			case 'F' :
				arguments.filename = optarg;
			break;

			case 'c' :
				arguments.comport = optarg;
			break;

			case 'e' :
				arguments.emulator = optarg;
			break;

			case 'f' :
				arguments.flash    = optarg;
			break;

			case 2:
			break;

			case 1:
				arguments.part = optarg;
			break;

			case 'D' :
			{
				char *p;
				arguments.chain.devbefore  = strtol(optarg, &p, 0);
				if (*p != ',') return -3;
				arguments.chain.devafter   = strtol(p+1, &p, 0);
				if (*p != ',') return -3;
				arguments.chain.bitsbefore = strtol(p+1, &p, 0);
				if (*p != ',') return -3;
				arguments.chain.bitsafter  = strtol(p+1, &p, 0);
				if (*p) return -3;
			}
			break;

			default:
				return -1;
		}
	}

	if (optind < argc)
		return -4;

	return 0;
}

int main(int argc, char **argv)
{
	parse_argv(argc, argv);

	if (arguments.version)
	{
		printf ("avr32gdbproxy 9.0.0\n");
		fflush(0);
		return 0;
	}

	if (arguments.help)
	{
		printf ("Help not implemented\n");
		return 1;
	}

	tcpstream tcps2;
	try { tcps2.openListen(arguments.traceaddress, TCP_NODELAY); } catch (const char *str) { printf("%s", str); return -1; }

	tcpstream tcps;
	try { tcps.openListen(arguments.gdbaddress, TCP_NODELAY); } catch (const char *str) { printf("%s", str); return -1; }
	fprintf(stderr, "%%ready\n");
	fflush(0);

	struct sockaddr_in ca2 = tcps2.waitForConnect();
	printf("Trace %s:%d connected\n", inet_ntoa(ca2.sin_addr), ntohs(ca2.sin_port));

	struct sockaddr_in ca = tcps.waitForConnect();
	printf("Debugger %s:%d connected\n", inet_ntoa(ca.sin_addr), ntohs(ca.sin_port));

	jtagAdapter *jtag = new AdapterXilinx_par();
	jtag->init();
	MCU_AVR32 *avr32Dev	= new MCU_AVR32(jtag);
	avr32Dev->init(0, 0);
	targetif *tif = new targetif(avr32Dev);
	tif->init();
	tif->stop();
	tif->check_stop();

	gdbbackend gdbback(&tcps, tif);
	gdbback.run();

	return 0;
}

