/** -*- c++ -*- **************************************************************
 **
 ** Copyright 2014 Daniel Paschka (daniel@drachenschuppe.de)
 **
 ** Licensed under the Apache License, Version 2.0 (the "License");
 ** you may not use this file except in compliance with the License.
 ** You may obtain a copy of the License at
 **
 ** http://www.apache.org/licenses/LICENSE-2.0
 **
 ** Unless required by applicable law or agreed to in writing, software
 ** distributed under the License is distributed on an "AS IS" BASIS,
 ** WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 ** See the License for the specific language governing permissions and
 ** limitations under the License.
 *****************************************************************************/

#define __STDC_FORMAT_MACROS
#include <inttypes.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <arpa/inet.h>

#include "../jtag/Device_AVR32.h"
#include "targetinterface.h"

const char *targetif::memory_map =
	"<?xml version=\"1.0\"?>\n"
	"<!DOCTYPE memory-map\n"
	"	PUBLIC \"+//IDN gnu.org//DTD GDB Memory Map V1.0//EN\"\n"
	"	\"http://sourceware.org/gdb/gdb-memory-map.dtd\">\n"
	"\n"
	"	<memory-map>\n"
	"		<memory type=\"ram\"   start=\"0x00000000\" length=\"0x00010000\"/>\n"
	/* flash memory maximum 2^32 = 8 MB*/
	"		<memory type=\"flash\" start=\"0x80000000\" length=\"0x00800000\">\n"
	"			<property name=\"blocksize\">0x200</property>\n"
	"		</memory>\n"
	/* 512 Bytes user flash page */
	"		<memory type=\"flash\" start=\"0x80800000\" length=\"0x00000200\">\n"
	"			<property name=\"blocksize\">0x200</property>\n"
	"		</memory>\n"
	"	</memory-map>\n"
	;

targetif::targetif(MCU_AVR32 *mcu)
{
	this->mcu = mcu;
	wait_stop = 0;
	last_stop_sig = 0x00;
	lastFlashPage = -1;
	dbgctrl = 0;
	memset (flashPage, 0xff, sizeof(flashPage));
	bp_addr = NULL;
	bpts = -1;
	wp_addr = NULL;
	wpts = -1;
}

targetif::~targetif()
{
	if (bp_addr) delete[] bp_addr;
	if (wp_addr) delete[] wp_addr;
}

int targetif::init()
{
	int ret;
	uint32_t val;

	ret = mcu->readNEXUS(MCU_AVR32::NEXUS_DID, val);
	if (ret < 0)
	{
		printf("Error accessing targetif device %d\n", ret);
		return -1;
	}

	if ((val & 0x3f) != 0x3f)
	{
		printf("Not an Atmel chip (ID=%08"PRIx32"), giving up!\n", val);
		return -2;
	}
	printf( "Product: %04"PRIx32", Rev %c\n", (val >> 12) & 0xffff, (char)('A' + (val >> 28)) );

	// activate debugging interface
	ret = setDC(MCU_AVR32::NEXUS_DC_DBE | MCU_AVR32::NEXUS_DC_DBR);
	if (ret < 0) return -3;
	wait_stop = 1;

	ret = mcu->readNEXUS(MCU_AVR32::NEXUS_NXCFG, val);
	bpts = (val>> 8) & 0xf;
	wpts = (val>>12) & 0xf;
	printf("%d HW breakpoints and %d watchpoints available\n", bpts, wpts);

	int i;
	bp_addr = new uint32_t[bpts];
	for (i=0; i<bpts; i++) bp_addr[i] = -1U;

	wp_addr = new uint32_t[wpts];
	for (i=0; i<wpts; i++) wp_addr[i] = -1U;

	cont();

	return 0;
}

void targetif::dumpMemory(int len, unsigned char *data)
{
	int i;
	for (i=0; i<len; i++)
	{
		printf("%02x ", data[i]);
		if ((i&15)==15)
			printf("\n");
	}

	if ((i&15)!=15)
		printf("\n");
}


int targetif::setDC(uint32_t val)
{
	if (dbgctrl == val)
		return 0;

	dbgctrl = val;

	return mcu->writeNEXUS(MCU_AVR32::NEXUS_DC, val);
}

int targetif::stop()
{
	setDC(MCU_AVR32::NEXUS_DC_DBE | MCU_AVR32::NEXUS_DC_DBR);
	wait_stop = 1;
	return 0;
}

int targetif::cont()
{
	setDC(MCU_AVR32::NEXUS_DC_DBE | MCU_AVR32::NEXUS_DC_IFM);
	mcu->writeNEXUS(MCU_AVR32::NEXUS_DINST, 0xd623d623);
	wait_stop = 1;
	return 0;
}

int targetif::step()
{
	setDC(MCU_AVR32::NEXUS_DC_DBE | MCU_AVR32::NEXUS_DC_SS);
	mcu->writeNEXUS(MCU_AVR32::NEXUS_DINST, 0xd623d623);
	wait_stop = 1;
	return 0;
}

int targetif::reset()
{
	mcu->haltCPU(0);
	setDC(MCU_AVR32::NEXUS_DC_DBE | MCU_AVR32::NEXUS_DC_RES | MCU_AVR32::NEXUS_DC_DBR);
	wait_stop = 1;
	return 0;
}

int targetif::check_stop()
{
	if (!wait_stop)
		return 0;

	uint32_t val;
	mcu->haltCPU(0);
	mcu->readNEXUS(MCU_AVR32::NEXUS_DS, val);
	if (val & MCU_AVR32::NEXUS_DS_DBS)
	{
		mcu->readNEXUS(MCU_AVR32::NEXUS_DPC, pc);
		wait_stop = 0;
		last_stop_sig = 0;

		if (val & MCU_AVR32::NEXUS_DS_DBA)
		{
			printf("Stopped @0x%08"PRIx32", reason was User break\n", pc);
			last_stop_sig = 2; // SIGINT
		}

		if (val & MCU_AVR32::NEXUS_DS_SSS)
		{
			printf("Stopped @0x%08"PRIx32", reason was Single Step\n", pc);
			last_stop_sig = 5; // SIGTRAP
		}

		if (val & MCU_AVR32::NEXUS_DS_SWB)
		{
			printf("Stopped reason was Software Breakpoint\n");
			last_stop_sig = 5; // SIGTRAP
		}

		if (val & MCU_AVR32::NEXUS_DS_HWB)
		{
			printf("Stopped @0x%08"PRIx32", reason was Hardware Breakpoint (Mask=%x)\n", pc, (uint8_t)((val>>8)&0xff) );
			last_stop_sig = 5; // SIGTRAP
		}

		if (val & MCU_AVR32::NEXUS_DS_HWE)
		{
			printf("Stopped reason was Hardware Error\n");
			last_stop_sig = 10; // SIGBUS
		}
		return 1;
	}

	return 0;
}

int targetif::hwpoint_set(int num, uint32_t *addrlist, uint32_t areg, uint32_t creg, uint32_t addr, int mode)
{
	int i;
	for (i=0; i<num; i++)
		if (addrlist[i] == -1U)
			break;
	if (i==num)
		return -2;

	int ret;
	ret = mcu->writeNEXUS(areg + i, addr);
	if (ret < 0)
		return -3;

	// enable hardware break/watch point
	ret = mcu->writeNEXUS(creg + i, (1<<30) | mode);
	if (ret < 0)
		return -4;
	addrlist[i] = addr;

	return 0;
}

int targetif::breakpoint_set(int type, uint32_t addr)
{
	printf ("TRGT: breakpoint set @0x%08"PRIx32", type %d\n", addr, type);

	// 0 = software breakpoint (memory breakpoint)
	// 1 = hardware breakpoint
	// 2 = write watchpoint (1)
	// 3 = read watchpoint  (0)
	// 4 = access watchpoint(2)

	switch (type)
	{
		case 1: return hwpoint_set(bpts, bp_addr, MCU_AVR32::NEXUS_BWA0A, MCU_AVR32::NEXUS_BWC0A, addr, 0);
		case 2: return hwpoint_set(wpts, wp_addr, MCU_AVR32::NEXUS_BWA3A, MCU_AVR32::NEXUS_BWC3A, addr, 1);
		case 3: return hwpoint_set(wpts, wp_addr, MCU_AVR32::NEXUS_BWA3A, MCU_AVR32::NEXUS_BWC3A, addr, 0);
		case 4: return hwpoint_set(wpts, wp_addr, MCU_AVR32::NEXUS_BWA3A, MCU_AVR32::NEXUS_BWC3A, addr, 2);

		// can't do other breakpoints
		default: return -1;
	}

	return 0;
}

int targetif::hwpoint_remove(int num, uint32_t *addrlist, uint32_t creg, uint32_t addr)
{
	// search address in list
	int i;
	for (i=0; i<num; i++)
		if (addrlist[i] == addr)
			break;
	// address no in list
	if (i==num)
		return -2;

	int ret = mcu->writeNEXUS(creg + i, 0); // disable breakpoint
	if (ret < 0)
		return -3;
	addrlist[i] = -1U;

	return 0;
}

int targetif::hwpoint_remove_all()
{
	for (int i=0; i<bpts; i++)
	{
		int ret = mcu->writeNEXUS(MCU_AVR32::NEXUS_BWC0A + i, 0); // disable breakpoint
		bp_addr[i] = -1U;
		if (ret < 0)
			return -1;
	}

	for (int i=0; i<wpts; i++)
	{
		int ret = mcu->writeNEXUS(MCU_AVR32::NEXUS_BWC3A + i, 0); // disable watchpoint
		wp_addr[i] = -1U;
		if (ret < 0)
			return -1;
	}
	return 0;
}

int targetif::breakpoint_remove(int type, uint32_t addr)
{
	printf ("TRGT: breakpoint remove @0x%08"PRIx32", type %d\n", addr, type);

	switch (type)
	{
		case 1: return hwpoint_remove(bpts, bp_addr, MCU_AVR32::NEXUS_BWC0A, addr);
		case 2:
		case 3:
		case 4: return hwpoint_remove(wpts, wp_addr, MCU_AVR32::NEXUS_BWC3A, addr);
		default: return -1;
	}

	return 0;
}

int targetif::read_system_register(int reg, uint32_t &val)
{
	// printf("TRGT: read system reg %x\n", reg);

	uint32_t back;
	read_mcu_register(0xe, back);
	mcu->writeNEXUS(MCU_AVR32::NEXUS_DINST,  0xe1be0000 | reg ); // mfsr r14, reg
	read_mcu_register(0xe, val);
	mcu->writeNEXUS(MCU_AVR32::NEXUS_DINST,  0xe06e0000 |  (back & 0x0000ffff)      ); // mov r14, LO
	mcu->writeNEXUS(MCU_AVR32::NEXUS_DINST,  0xea1e0000 | ((back & 0xffff0000)>>16) ); // mov r14, HI

	return 0;
}

int targetif::write_system_register(int reg, uint32_t val)
{
	// printf("TRGT: write system reg %d = %x\n", reg, val);

	uint32_t back;
	read_mcu_register(0xe, back);

	mcu->writeNEXUS(MCU_AVR32::NEXUS_DINST,  0xe06e0000 |  (val & 0x0000ffff)      ); // mov r14, LO
	mcu->writeNEXUS(MCU_AVR32::NEXUS_DINST,  0xea1e0000 | ((val & 0xffff0000)>>16) ); // mov r14, HI
	mcu->writeNEXUS(MCU_AVR32::NEXUS_DINST,  0xe3be0000 | reg); // mtsr reg, r14
	mcu->writeNEXUS(MCU_AVR32::NEXUS_DINST,  0xe06e0000 |  (back & 0x0000ffff)      ); // mov r14, LO
	mcu->writeNEXUS(MCU_AVR32::NEXUS_DINST,  0xea1e0000 | ((back & 0xffff0000)>>16) ); // mov r14, HI

	return 0;
}

int targetif::read_mcu_register(int reg, uint32_t &val)
{
	// PC needs special handling
	if (reg == 0x0f)
	{
		uint32_t back;
		read_mcu_register(0xe, back);

		mcu->writeNEXUS(MCU_AVR32::NEXUS_DINST,  0xe1be0014 ); // mfsr r14, RAR_DBG
		read_mcu_register(0xe, val);

		mcu->writeNEXUS(MCU_AVR32::NEXUS_DINST,  0xe06e0000 |  (back & 0x0000ffff)      ); // mov r14, LO
		mcu->writeNEXUS(MCU_AVR32::NEXUS_DINST,  0xea1e0000 | ((back & 0xffff0000)>>16) ); // mov r14, HI
	}
	else
	{
		mcu->writeNEXUS(MCU_AVR32::NEXUS_DINST, 0xe7b00044 | (reg<<16) ); // DCCPU
		mcu->readNEXUS(MCU_AVR32::NEXUS_DCCPU, val);
	}
	return 0;
}

int targetif::write_mcu_register(int reg, uint32_t val)
{
	printf ("TRGT: write 0x%08"PRIx32" to register %d\n", val, reg);

	// MCU has 16 registers
	if (reg >= 0x10)
		return -1;

	if ( (reg == 0x0f) && (pc != val))
	{
		// printf("Write to PC, changing return address\n");
		uint32_t back;

		read_mcu_register(0xe, back);
		mcu->writeNEXUS(MCU_AVR32::NEXUS_DINST,  0xe06e0000 |  (val & 0x0000ffff)      ); // mov r14, LO
		mcu->writeNEXUS(MCU_AVR32::NEXUS_DINST,  0xea1e0000 | ((val & 0xffff0000)>>16) ); // mov r14, HI
		mcu->writeNEXUS(MCU_AVR32::NEXUS_DINST,  0xe3be0014 ); // mtsr RAR_DBG, r14
		mcu->writeNEXUS(MCU_AVR32::NEXUS_DINST,  0xe06e0000 |  (back & 0x0000ffff)      ); // mov r14, LO
		mcu->writeNEXUS(MCU_AVR32::NEXUS_DINST,  0xea1e0000 | ((back & 0xffff0000)>>16) ); // mov r14, HI
		pc = val;
	}
	else
	{
		mcu->writeNEXUS(MCU_AVR32::NEXUS_DINST,  0xe0600000 | (reg<<16) |  (val & 0x0000ffff)      ); // mov r?, LO
		mcu->writeNEXUS(MCU_AVR32::NEXUS_DINST,  0xea100000 | (reg<<16) | ((val & 0xffff0000)>>16) ); // mov r?, HI
	}

	return 0;
}

int targetif::memory_read(uint32_t addr, int len, unsigned char *data)
{
	// printf ("TRGT: memory read %d bytes @0x%08"PRIx32":\n", len, addr);

	// read unaligned data at begining
	while ((addr&3) && len)
	{
		uint32_t val;
		mcu->readAddrSingle(addr, MCU_AVR32::Byte, val);
		*data = val >> (24 - 8*(addr & 3)) & 0xff;
		// printf("single byte read @%08x\n, %08x %02x\n", addr, val, *data);
		data++;
		addr++;
		len--;
	}

	// read aligned data block
	uint32_t *data32 = (uint32_t*)data;
	mcu->readAddrBlock(addr, data32, len/4);
	int i;
	for (i=0; i<len/4; i++)
		data32[i] = htonl(data32[i]);

	data += i*4;
	addr += i*4;
	len  -= i*4;

	// read unaligned data at the end
	while (len)
	{
		uint32_t val;
		mcu->readAddrSingle(addr, MCU_AVR32::Byte, val);
		*data = val >> (24 - 8*(addr & 3)) & 0xff;
		// printf("single byte read @%08x\n, %08x %02x\n", addr, val, *data);
		data++;
		addr++;
		len--;
	}

	return 0;
}

int targetif::memory_write(uint32_t addr, int len, unsigned char *data)
{
	// printf ("TRGT: memory write %d bytes @0x%08"PRIx32":\n", len, addr);
	dumpMemory(len, data);

	// write unaligned data at begining
	while ((addr&3) && len)
	{
		uint32_t val = (*data) << (24 - 8*(addr & 3));
		mcu->writeAddrSingle(addr, MCU_AVR32::Byte, val);
		data++;
		addr++;
		len--;
	}

	// write aligned data block
	uint32_t *data32 = (uint32_t*)data;
	int i;
	for (i=0; i<len/4; i++)
		data32[i] = htonl(data32[i]);
	mcu->writeAddrBlock(addr, data32, len/4);

	data += i*4;
	addr += i*4;
	len  -= i*4;

	// write unaligned data at end
	while (len)
	{
		uint32_t val = (*data) << (24 - 8*(addr & 3));
		mcu->writeAddrSingle(addr, MCU_AVR32::Byte, val);
		data++;
		addr++;
		len--;
	}

	return 0;
}

int targetif::flashErase(uint32_t addr, int len)
{
	printf ("TRGT: flasherase of %d bytes @0x%08"PRIx32":\n", len, addr);
	if (addr < 0x80000000)
		return -1;

	mcu->haltCPU(1);
	while (len)
	{
		int page = (addr-0x80000000) / 0x200;
		printf ("TRGT: eraseing page %d @0x%08"PRIx32":\n", page, addr);
		mcu->flashErasePage(page);
		len  -= 0x200;
		addr += 0x200;
	}

	return 0;
}

int targetif::flashWrite(uint32_t addr, int len, unsigned char *data)
{
	printf ("TRGT: flashwrite of %d bytes @0x%08"PRIx32"\n", len, addr);
	if (addr < 0x80000000)
		return -1;

	//dumpMemory(len, data);
	mcu->haltCPU(1);
	int i;
	for (i=0; i<len; i++)
	{
		int page = (addr+i-0x80000000) / 512;
		if (lastFlashPage == -1) lastFlashPage = page;

		if (lastFlashPage != page)
			flashPageFinish();

		flashPage[(addr + i) & 0x1ff] = data[i];
		lastFlashPage = page;
	}

	return 0;
}

void targetif::flashPageFinish()
{
	if (lastFlashPage < 0)
		return;

	int i;
	uint32_t *flashPage32 = (uint32_t *)flashPage;
	for (i=0; i<512/4; i++)
		flashPage32[i] = htonl(flashPage32[i]);

	printf ("TRGT: flashwriting page @0x%08x\n", 0x80000000 + lastFlashPage*512);
	mcu->flashWriteWords(0x80000000 + lastFlashPage*512, flashPage32, 512/4);
	memset (flashPage, 0xff, sizeof(flashPage));
	lastFlashPage = -1;
}

void targetif::flashFinish()
{
	flashPageFinish();
	mcu->haltCPU(0);
	dbgctrl = 0;
	stop();
}

