/** -*- c++ -*- **************************************************************
 **
 ** Copyright 2014 Daniel Paschka (daniel@drachenschuppe.de)
 **
 ** Licensed under the Apache License, Version 2.0 (the "License");
 ** you may not use this file except in compliance with the License.
 ** You may obtain a copy of the License at
 **
 ** http://www.apache.org/licenses/LICENSE-2.0
 **
 ** Unless required by applicable law or agreed to in writing, software
 ** distributed under the License is distributed on an "AS IS" BASIS,
 ** WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 ** See the License for the specific language governing permissions and
 ** limitations under the License.
 *****************************************************************************/

#include <stdio.h>
#include <string.h>

#include "tcpstream.h"

#ifdef __WIN32__
static WSADATA wsaData;
static int     wsaInitalized = 0;
#endif // __WIN32__

tcpstreambuf::tcpstreambuf() : _putBackCnt(0)
{
#ifdef __WIN32__
	if (!wsaInitalized)
	{
		// Initialize Winsock
		int iResult = WSAStartup(MAKEWORD(2,2), &wsaData);
		if (iResult != NO_ERROR)
		{
			fprintf(stderr, "WSAStartup failed: %d\n", iResult);
			return;
		}
		wsaInitalized = 1;
		fprintf(stderr, "WSAStartup succeeded\n");
	}
#endif // __WIN32__

	_clientfd = -1;
	_serverfd = -1;
}

void tcpstreambuf::openListen(uint16_t port, int socketOption)
{
	_serverfd = -1;

#ifdef __WIN32__
	if (!wsaInitalized)
		throw "WSA initialization required\n";
#endif // __WIN32__

	// open TCP socket
	_serverfd = socket(AF_INET, SOCK_STREAM, 0);
	if ( _serverfd < 0 )
		throw "error initalizing socket\n";

	int flag = 1;
	int result = setsockopt
	(
		_serverfd,     /* socket affected */
		IPPROTO_TCP,   /* set option at TCP level */
		socketOption,  /* name of option */
		(char*)&flag,  /* the cast is historical cruft */
		sizeof(int)    /* length of option value */
	);
	if (result < 0)
		throw "error setting TCP flags\n";

	// prepare socket
	memset(&_serverAddr, 0, sizeof(_serverAddr));
	_serverAddr.sin_family = AF_INET;
	_serverAddr.sin_port = htons(port);
	_serverAddr.sin_addr.s_addr = INADDR_ANY;
	if ( bind(_serverfd, (struct sockaddr*)&_serverAddr, sizeof(_serverAddr)) < 0 )
		throw "error binding socket\n";

	// wait for incoming connection
	if ( listen(_serverfd, 1) < 0 )
		throw "error on start listening\n";
}

struct sockaddr_in tcpstreambuf::waitForConnect()
{
	if (_serverfd < 0)
	{
		_clientfd = -1;
		return _clientAddr;
	}

	socklen_t addrlen = sizeof(_clientAddr);
	_clientfd = accept(_serverfd, (struct sockaddr*)&_clientAddr, &addrlen);
	return _clientAddr;
}

std::streamsize tcpstreambuf::showmanyc()
{
	struct timeval timeout;
	timeout.tv_sec  = 0;
	timeout.tv_usec = 1000; // 1 ms

	fd_set f;
	FD_ZERO(&f);
	FD_SET((unsigned)_clientfd, &f);
	int ret = select(_clientfd+1, &f, NULL, NULL, &timeout);

	// select returned so at least one char is present
	if (ret > 0)
		return 1;

	return 0;
}

int tcpstreambuf::underflow()
{
	// printf("underflow called %p %p %p\n", gptr(), eback(), egptr());

	if (gptr() < egptr()) // buffer not exhausted
		return traits_type::to_int_type(*gptr());

	char *start = _buffer;

	if (eback() == _buffer) // true when this isn't the first fill
	{
		// Make arrangements for putback characters
		memmove(_buffer, egptr() - _putBackCnt, _putBackCnt);
		start += _putBackCnt;
	}

	// we are not connected there is nothing to read
	if (_clientfd < 0)
		return traits_type::eof();

	// read the data and put it into the buffer
	int ret = recv(_clientfd, start, sizeof(_buffer) - (start-_buffer), 0);
	if (ret <= 0)
		return traits_type::eof();

	setg(_buffer, start, start + ret);

	return traits_type::to_int_type(*gptr());
}

std::streamsize tcpstreambuf::xsputn(const char* s, std::streamsize n)
{
	// we are not connected just throw the data away
	if (_clientfd < 0)
		return n;

	// printf("xsputn called %p, %d\n", s, n);
	send(_clientfd, s, n, 0);
	return n;
}

