/** -*- c++ -*- **************************************************************
 **
 ** Copyright 2014 Lynx Technik AG (www.lynx-technik.com)
 **
 ** Licensed under the Apache License, Version 2.0 (the "License");
 ** you may not use this file except in compliance with the License.
 ** You may obtain a copy of the License at
 **
 ** http://www.apache.org/licenses/LICENSE-2.0
 **
 ** Unless required by applicable law or agreed to in writing, software
 ** distributed under the License is distributed on an "AS IS" BASIS,
 ** WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 ** See the License for the specific language governing permissions and
 ** limitations under the License.
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <getopt.h>

#include "elfread.h"
#include "../jtag/jtagAdapter.h"
#include "../jtag/Device_AVR32.h"

#include "../jtag/AdapterXilinxPar.h"

enum eCmd
{
	eCmd_none,

	eCmd_chiperase,
	eCmd_erase,
	eCmd_program,
	eCmd_read,
	eCmd_verify,
	eCmd_run,
	eCmd_reset,
	eCmd_halt,
	eCmd_cpuinfo,
	eCmd_readregs,

	eCmd_scanchain,
};

struct args_st
{
	unsigned help           :1;
	unsigned list           :1;
	unsigned version        :1;
	unsigned gui            :1;

	char     *port;
	char     *programmer;
	unsigned  verbose_level;

	char     *part;

	struct
	{
		int devbefore;
		int devafter;
		int bitsbefore;
		int bitsafter;
	} chain;

	struct prog_st
	{
		unsigned  verify    : 1;
		unsigned  erase     : 1;
		unsigned  chiperase : 1;
		unsigned  reset     : 1;
		unsigned  run       : 1;
		unsigned  nounlock  : 1;
		unsigned  format    : 2;
		unsigned  clockSrc  : 2;
		unsigned  offset;
		unsigned  size;
		char     *flash;
		char     *file;
	} prog;

	enum eCmd cmd;
} arguments;


int parse_argv_program(int argc, char **argv)
{
	static struct option long_options[] =
	{
		{"format",              required_argument, 0, 'F'},
		{"offset",              required_argument, 0, 'O'},
		{"size",                required_argument, 0, 's'},
		{"verify",              no_argument,       0, 'v'},
		{"chiperase",           no_argument,       0, 'E'},
		{"erase",               no_argument,       0, 'e'},
		{"flash",               required_argument, 0, 'f'},
		{"reset",               no_argument,       0, 'R'},
		{"run",                 optional_argument, 0, 'r'},
		{"clock",               required_argument, 0, 'c'},
		{"no-unlock",           no_argument,       0, 'u'},
	};

	int c;
	while (1)
	{
		/* getopt_long stores the option index here. */
		int option_index = -1;

		c = getopt_long (argc, argv, "vEeRuF:O:s:f:c:r::", long_options, &option_index);

		/* Detect the end of the options. */
		if (c == -1)
			break;

		switch (c)
		{
			case 'r' : arguments.prog.run = 1; break;
			case 'R' : arguments.prog.reset = 1; break;
			case 'E' : arguments.prog.chiperase = 1; break;
			case 'e' : arguments.prog.erase = 1; break;
			case 'u' : arguments.prog.nounlock = 1; break;
			case 'v' : arguments.prog.verify = 1; break;
			case 'f' : arguments.prog.flash = strdup(optarg); break;
			case 'O' : arguments.prog.offset = strtol(optarg, NULL, 0);	break;
			case 'c' :
				if (!strcmp(optarg, "int"))    arguments.prog.clockSrc = 0;
				if (!strcmp(optarg, "xtal"))   arguments.prog.clockSrc = 1;
				if (!strcmp(optarg, "extclk")) arguments.prog.clockSrc = 2;
			break;

			case 'F' :
				if (!strcmp(optarg, "elf")) arguments.prog.format = 0;
				if (!strcmp(optarg, "bin")) arguments.prog.format = 1;
			break;

			default:
				printf("XX %d %c\n",c , c);
		}
	}

	if (optind >= argc)
	{
		printf("program: missing file argument.\n");
		return -1;
	}

	arguments.prog.file = strdup(argv[optind]);
	
	return 0;
}

int parse_argv(int argc, char **argv)
{
	int c;
	memset (&arguments, 0, sizeof(arguments));

	while (1)
	{
		static struct option long_options[] =
		{
			{"help",                no_argument,       0, 'h'},
			{"list",                no_argument,       0, 'l'},
			{"verbose",             optional_argument, 0, 'v'},
			{"version",             no_argument,       0, 'V'},
			{"comport",             required_argument, 0, 'c'},
			{"xmlpath",             required_argument, 0, 'x'},
			{"programmer",          required_argument, 0, 'p'},
			{"gui",                 no_argument,       0, 'G'},
			{"daisy",               required_argument, 0, 'D'},
			{"part",                required_argument, 0, 1  },
			{"no-tool-setup-check", required_argument, 0, 2  },
		};

		/* getopt_long stores the option index here. */
		int option_index = -1;

		c = getopt_long (argc, argv, "+hlVGv::c:x:p:D:", long_options, &option_index);

		/* Detect the end of the options. */
		if (c == -1)
			break;

		switch (c)
		{
			case 'h' : arguments.help = 1; break;
			case 'l' : arguments.list = 1; break;
			case 'V' : arguments.version = 1; break;
			case 'G' : arguments.gui = 1; break;
			case 'c' : arguments.port = strdup(optarg); break;
			case 'p' : arguments.programmer = strdup(optarg); break;
			case 1   : arguments.part = strdup(optarg); break;

			case 'v' : arguments.verbose_level = 1;
				if (optarg)
				{
					char *p, l = strlen(optarg);
					arguments.verbose_level = strtol(optarg, &p, 0);

					if (p - optarg != l)
					{
						printf("This \"%s\" is not a number!\n", optarg);
						return -1;
					}
				}
			break;

			case 'D' :
			{
				char *p;
				arguments.chain.devbefore  = strtol(optarg, &p, 0);
				if (*p != ',') return -3;
				arguments.chain.devafter   = strtol(p+1, &p, 0);
				if (*p != ',') return -3;
				arguments.chain.bitsbefore = strtol(p+1, &p, 0);
				if (*p != ',') return -3;
				arguments.chain.bitsafter  = strtol(p+1, &p, 0);
				if (*p) return -3;
			}
			break;

			default:
				return -1;
		} // switch
	} // while (1)

	if (optind >= argc)
	{
		printf("Missing command. Try avr32program -h for help.");
		return -1;
	}

	char *cmd = argv[optind++];
	if (!strcmp(cmd, "chiperase")) { arguments.cmd = eCmd_chiperase; return 0; }
	if (!strcmp(cmd, "erase"))     { arguments.cmd = eCmd_erase; return 0; }
	if (!strcmp(cmd, "program"))   { arguments.cmd = eCmd_program; return parse_argv_program(argc, argv); }
	if (!strcmp(cmd, "read"))      { arguments.cmd = eCmd_read; return 0; }
	if (!strcmp(cmd, "verify"))    { arguments.cmd = eCmd_verify; return 0; }
	if (!strcmp(cmd, "halt"))      { arguments.cmd = eCmd_halt; return 0; }
	if (!strcmp(cmd, "run"))       { arguments.cmd = eCmd_run; return 0; }
	if (!strcmp(cmd, "reset"))     { arguments.cmd = eCmd_reset; return 0; }
	if (!strcmp(cmd, "cpuinfo"))   { arguments.cmd = eCmd_cpuinfo; return 0; }
	if (!strcmp(cmd, "readregs"))  { arguments.cmd = eCmd_readregs; return 0; }
	if (!strcmp(cmd, "scanchain")) { arguments.cmd = eCmd_scanchain; return 0; }

	printf("Command %s not understood\n", cmd);
	return -1;
}

#if 0
./avr32program -c USB:0 -pjtagicemkii --part UC3B064 program -finternal@0x80000000,512Kb -cxtal -e -v -O0x80000000 -Fbin /tmp/at32uc3b-isp-1.0.3.bin2850674796598394311.bin
./avr32program -c USB:0 -pjtagicemkii --part UC3B064 program -finternal@0x80000000,512Kb -cxtal -e -v -O0x808001FC -Fbin /tmp/userConf2039445217338415070,dat
./avr32program -c USB:0 -pjtagicemkii --part UC3B064 run
./avr32program -c USB:0 -pjtagicemkii --part UC3B064 reset
./avr32program -c USB:0 -pjtagicemkii --gui --part UC3B064 program -finternal@0x80000000 -cint -F elf -O 0 -r /home/daniel/Documents/bs571800/emb32/src/app/bs571800/BS571800_T.elf
./avr32program -c USB:0 -pjtagicemkii --gui --part UC3B064 read -o /home/daniel/Documents/a -s 65536 -O 0 -F bin
./avr32program -c USB:0 -pjtagicemkii --gui --part UC3B064 verify -O 0 -F bin /home/daniel/Documents/bs571800/emb32/src/app/bs571800/BS571800_T.elf
./avr32program -c USB:0 -pjtagicemkii --gui --part UC3B064 erase -finternal@0x80000000
./avr32program -c USB:0 -pjtagicemkii --gui --part UC3B064 erase -finternal@0x80000000 --also-user-page
./avr32program -c USB:0 -pjtagicemkii --gui --part UC3B064 erase -finternal@0x80000000 --only-user-page
./avr32program -c USB:0 -pjtagicemkii --gui --part UC3B064 chiperase
./avr32program -c USB:0 -pjtagicemkii --gui --part UC3B064 readregs
./avr32program -c USB:0 -pjtagicemkii --gui cpuinfo --features

#endif
void showHelp(void)
{
	printf("  -h, --help			Display this help and exit.\n");
	printf("  -l, --list			Detect and print information about available\n");
	printf("				USB programmers.\n");
	printf("  -v, --verbose <level>		Display diagnostic messages.\n");
	printf("  -V, --version			Print the version number.\n");
	printf("  -c, --comport <port>		Connect to given serial port. If <port> is\n");
	printf("				'USB', connect to a USB device. If <port> is\n");
	printf("				'USB:serial', the programmer with the given\n");
	printf("				serial number will be used. The default is to\n");
	printf("				connect to a USB device.\n");
	printf("  -x, --xmlpath <directory>	Specify path to XML part description files.\n");
	printf("  -p, --programmer <programmer>	Specify which kind of programmer to use.\n");
	printf("				Can be either jtagicemkii (the default),\n");
	printf("				stk600, qt600, avrdragon or avrone.\n");
	printf("  -G, --gui			Indicates that program is being called from a\n");
	printf("				graphical user interface. Changes progress\n");
	printf("				reporting.\n");
	printf("  -D, --daisy <devbefore,devafter,bitsbefore,bitsafter>\n");
	printf("				Specify JTAG daisy-chain parameters.\n");
	printf("  --part <partname>		Specify part name, fail unless target matches.\n");
	printf("  --no-tool-setup-check		Prevent the programmer from checking its\n");
	printf("				setup against the selected part. Not applicable\n");
	printf("				to all tools.\n");
}

int scanChain(jtagAdapter *jtag)
{
	bitvector r;
	shiftData d;
	d.length = 32;

	jtag->doState(jtagAdapter::RESET);

	do
	{
		r.clear();
		jtag->shift(jtagAdapter::DRSHIFT, &d, jtagAdapter::DRSHIFT, &r);
		uint32_t id = r.toNum();
		if ( (id == 0xFFFFffff) || !(id&1) )
		{
			printf("\n");
			return 0;
		}
		printf("chain: %08x\t", id);
	} while (1);

	return 1;
}

int program(MCU_AVR32 *avr32Dev)
{
	avr32Dev->init(1, !arguments.prog.nounlock);

	printf("programming file: %s\n", arguments.prog.file);
	int fd = open(arguments.prog.file, 0);
	if (fd < 0)
		return -1;

	struct page_st *pages = ELFread(fd);
	int pagesTodo = 0;
	for (int p=0; p<1024; p++)
		if (pages[p].used)
			pagesTodo++;

	if (arguments.prog.erase)
	{
		int pagesDone = 0;
		for (int p=0; p<1024; p++)
			if (pages[p].used)
			{
				avr32Dev->flashErase(0x80000000 + PAGESIZE*p);
				pagesDone++;
				fprintf(stderr, "%%%.1f Erasing flash\n", 100.0*pagesDone/pagesTodo);
			}
	}

	pagesTodo*=2;
	int pagesDone = 0;
	{
		for (int p=0; p<1024; p++)
			if (pages[p].used)
			{
				int ret = avr32Dev->flashWrite(0x80000000 + PAGESIZE*p, PAGESIZE, pages[p].data);
				if (ret < 0)
				{
					fprintf(stderr, "Flashwrite failed with error %d\n", ret);
					return -1;
				}
				pagesDone++;
				fprintf(stderr, "%%%.1f Programming flash\n", 100.0*pagesDone/pagesTodo);
			}
	}

	if (arguments.prog.verify)
	{
		for (int p=0; p<1024; p++)
			if (pages[p].used)
			{
				uint8_t data[PAGESIZE];
				int ret = avr32Dev->flashRead(0x80000000 + PAGESIZE*p, PAGESIZE, data);
				if (ret < 0)
				{
					fprintf(stderr, "Flashread failed with error %d\n", ret);
					return -1;
				}
				if (memcmp(data, pages[p].data, PAGESIZE))
				{
					fprintf(stderr, "Verify failed\n");
					return 1;
				}
				pagesDone++;
				fprintf(stderr, "%%%.1f Verifing flash\n", 100.0*pagesDone/pagesTodo);
			}
	}
	fprintf(stderr, "%%done\n");

	if (arguments.prog.reset)
		avr32Dev->resetCPU(1);

	if (arguments.prog.run)
		avr32Dev->haltCPU(0);
	
	return 0;
}

int main(int argc, char **argv)
{
	int ret = parse_argv(argc, argv);
	if (ret < 0)
		return 1;

	if (arguments.help)
	{
		showHelp();
		return 1;
	}

//	fprintf(stderr, "Connected to JTAGICE mkII version 6.6, 6.6 at USB:090000000000.\n\r");
	jtagAdapter *jtag = new AdapterXilinx_par();
	jtag->init();

	if (arguments.cmd == eCmd_scanchain)
		return scanChain(jtag);

	MCU_AVR32 *avr32Dev	= new MCU_AVR32(jtag);
	if (arguments.cmd == eCmd_program)
		return program(avr32Dev);

	avr32Dev->init(1, 0);
	if (arguments.cmd == eCmd_reset)
	{
		avr32Dev->resetCPU(3);
		avr32Dev->haltCPU(0);
	}

	if (arguments.cmd == eCmd_halt)
		avr32Dev->haltCPU(1);

	if (arguments.cmd == eCmd_run)
		avr32Dev->haltCPU(0);

	return 0;
}

