/** -*- c++ -*- **************************************************************
 **
 ** Copyright 2014 Lynx Technik AG (www.lynx-technik.com)
 **
 ** Licensed under the Apache License, Version 2.0 (the "License");
 ** you may not use this file except in compliance with the License.
 ** You may obtain a copy of the License at
 **
 ** http://www.apache.org/licenses/LICENSE-2.0
 **
 ** Unless required by applicable law or agreed to in writing, software
 ** distributed under the License is distributed on an "AS IS" BASIS,
 ** WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 ** See the License for the specific language governing permissions and
 ** limitations under the License.
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include "parallelPort.h"
#include "AdapterXilinxPar.h"

AdapterXilinx_par::AdapterXilinx_par() : jtagAdapter()
{
	try
	{
		par = new parallelPort(0);
	}
	catch (char *e)
	{
		free (e);
		return;
	}

    curState = OFF;
    usdelay = 0;
}

AdapterXilinx_par::~AdapterXilinx_par()
{
    if (curState > OFF)
        release();
    if(par)
        delete par;
}

int AdapterXilinx_par::init()
{
    unsigned char v00, v40;

    // didn't even pass the constructor
    if (curState == FAIL)
    {
        printf("Xilinx Adapter could not be initialized.\n");
        return -1;
    }

    out(0x00);
    v00 = in();

    out(0x40);
    v40 = in();

    // adapter connected? D6->PaperEmpty
    if ( (v00 & 0x20) || (!(v40 & 0x20)) )
    {
        printf("Xilinx Adapter not found.\n");
        return -3;
    }

    // power sense pin of adapter high?
    if ( !(v00 & (1<<_PWR)) )
    {
        printf("Please connect adapter to targetboard. (no power)\n");
        return -1001;
    }

    // set to state initalized
    curState = INIT;
    portval  = 0;

    // set adapter to programming mode
    portval  = 1 <<_PROG;
    out (portval);

    return 0;
}

void AdapterXilinx_par::release(bool bFullRelease)
{
    if (curState <= OFF)
        return;

    doState(RESET);

    portval &= ~(1<<_PROG);
    out (portval);
    curState = OFF;
}

void AdapterXilinx_par::setJTAGSpeed(int Hz)
{
    float us = 1000000.0/Hz - 4;
    us *= 512;
    usdelay = (int)us;
    if (usdelay < 0)
        usdelay = 0;

    printf("using extradelay of %d us every 512 bit\n", usdelay);
}

void AdapterXilinx_par::setTCK(int val)
{
    static int bitCnt = 0;
    if (val)
    {
        if (!(portval&(1<<_TCK)))
        {
            // add an additional pause
            bitCnt ++;
            if ( usdelay && (bitCnt >= 512) )
            {
                bitCnt = 0;
                usleep(usdelay);
            }
        }

        portval |=   1<<_TCK;
    }
    else
        portval &= ~(1<<_TCK);
    out (portval);
}

void AdapterXilinx_par::setTMS(int val)
{
    if (val)
        portval |=   1<<_TMS;
    else
        portval &= ~(1<<_TMS);

    out (portval);
}

void AdapterXilinx_par::setTDI(int val)
{
    if (val)
        portval |=   1<<_TDI;
    else
        portval &= ~(1<<_TDI);

    out (portval);
}

int AdapterXilinx_par::getTDO()
{
    unsigned char val = in();
    return (val >> _TDO) & 1;
}

void AdapterXilinx_par::out(unsigned char val)
{
    static unsigned char last = 0xff;
    if (last == val)
        return;

    last = val;
    par->data_out (val);
}

char AdapterXilinx_par::in()
{
    return par->status_in();
}

void AdapterXilinx_par::clock(int val)
{
    for (int i=0;i<val;i++)
    {
        setTCK(1);
        setTCK(0);
    }
}

// switch to the requested state
void AdapterXilinx_par::doState(int val)
{
    if (curState == OFF)
        goto out;

    if (val == RESET)
    {
        setTMS(1);
        clock(5);
        curState = RESET;
        goto out;
    }

    while (curState != val)
    {
        switch (curState)
        {
            case INIT:
                doState(RESET);
                break;

            case RESET:
            {
                setTMS(0); clock(1); curState = IDLE;
            }
            break;

            case IDLE:
                setTMS(1); clock(1); curState = DRSELECT;
                break;

                // DR states
            case DRSELECT:
                if (val >= IRSELECT)
                { setTMS(1); clock(1); curState = IRSELECT; }
                else
                { setTMS(0); clock(1); curState = DRCAPTURE; }
                break;

            case DRCAPTURE:
                if (val == DRSHIFT)
                { setTMS(0); clock(1); curState = DRSHIFT; }
                else
                { setTMS(1); clock(1); curState = DREXIT1; }
                break;

            case DRSHIFT:
                setTMS(1); clock(1); curState = DREXIT1;
                break;

            case DREXIT1:
                if (val == DRPAUSE)
                { setTMS(0); clock(1); curState = DRPAUSE; }
                else
                { setTMS(1); clock(1); curState = DRUPDATE; }
                break;

            case DRPAUSE:
                setTMS(1); clock(1); curState = DREXIT2;
                break;

            case DREXIT2:
                if (val == DRSHIFT)
                { setTMS(0); clock(1); curState = DRSHIFT; }
                else
                { setTMS(1); clock(1); curState = DRUPDATE; }
                break;

            case DRUPDATE:
                if (val == IDLE)
                { setTMS(0); clock(1); curState = IDLE; }
                else
                { setTMS(1); clock(1); curState = DRSELECT; }
                break;

                // IR states
            case IRSELECT:
                setTMS(0); clock(1); curState = IRCAPTURE;
                break;

            case IRCAPTURE:
                if (val == IRSHIFT)
                { setTMS(0); clock(1); curState = IRSHIFT; }
                else
                { setTMS(1); clock(1); curState = IREXIT1; }
                break;

            case IRSHIFT:
                setTMS(1); clock(1); curState = IREXIT1;
                break;

            case IREXIT1:
                if (val == IRPAUSE)
                { setTMS(0); clock(1); curState = IRPAUSE; }
                else
                { setTMS(1); clock(1); curState = IRUPDATE; }
                break;

            case IRPAUSE:
                setTMS(1); clock(1); curState = IREXIT2;
                break;

            case IREXIT2:
                if (val == IRSHIFT)
                { setTMS(0); clock(1); curState = IRSHIFT; }
                else
                { setTMS(1); clock(1); curState = IRUPDATE; }
                break;

            case IRUPDATE:
                if (val == IDLE)
                { setTMS(0); clock(1); curState = IDLE; }
                else
                { setTMS(1); clock(1); curState = DRSELECT; }
                break;

                // should not happen
            default:
                return;
        }
    }

  out:
	;
}

int AdapterXilinx_par::shift (
    int shiftState, struct shiftData *dat, int endState, bitvector *readback
    )
{

    bitvector back;
    unsigned before = (shiftState == IRSHIFT) ? chain.instBefore : chain.devBefore;
    unsigned after  = (shiftState == IRSHIFT) ? chain.instAfter  : chain.devAfter;
    int first = shiftState != curState;
    int last  = endState   != shiftState;

    if (curState == OFF)
        return -2;

    // switch to shifting state. should be IRSHIFT or DRSHIFT
    doState(shiftState);

    // no readback, activate verifybuffer
    if (!readback && dat->tdo.size())
        readback = &back;

    // take care of the devices before our target in the jtag chain
    setTDI(1);
    if (first)
        clock(before);

    // shift out our data
    for (unsigned i=0; i<dat->length; i++)
    {
        // together with the last bit we need to change the state
        if ( (i==dat->length-1) && (!after) && last)
            setTMS(1);

        if (i < dat->tdi.size())
            setTDI(dat->tdi[dat->tdi.size()-1-i]);

        if (readback)
            readback->push_back(getTDO());

        clock(1);
    }

    //  take care of the devices after our target in the jtag chain
    if (after && last)
    {
        setTDI(1);
        clock (after-1);
        setTMS(1);
        clock(1);
    }
    setTDI(1);

    // update state variable
    if (last)
        curState++;

    // we have data to verify
    if (dat->tdo.size())
    {
        int ret = checkReadbackData(dat->length, &dat->tdo, readback, &dat->mask);
        if (ret)
            return ret;
    }

    // switch to the requested endState
    doState(endState);
    return 0;
}

