/** -*- c++ -*- **************************************************************
 **
 ** Copyright 2014 Daniel Paschka (daniel@drachenschuppe.de)
 **
 ** Licensed under the Apache License, Version 2.0 (the "License");
 ** you may not use this file except in compliance with the License.
 ** You may obtain a copy of the License at
 **
 ** http://www.apache.org/licenses/LICENSE-2.0
 **
 ** Unless required by applicable law or agreed to in writing, software
 ** distributed under the License is distributed on an "AS IS" BASIS,
 ** WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 ** See the License for the specific language governing permissions and
 ** limitations under the License.
 *****************************************************************************/

#ifdef __WIN32__
#	include <winsock2.h>
#else
#	include <arpa/inet.h>
#endif // __WIN32__

#include <string.h>
#include <stdio.h>
#include <sys/time.h>

#include "bitvector.h"
#include "jtagAdapter.h"
#include "Device_AVR32.h"

#define SAB_HSB 0x04
#define SAB_OCD 0x01

void _usleep(int us) {};

int MCU_AVR32::setCommand(int cmd)
{
    int rc;
	if ((cmd == lastCommand) && (jtag->getState() != jtagAdapter::RESET))
		return 0;

	shiftData shcmd;
	shcmd.length = 5;
	shcmd.tdi.push_back_value(cmd, 5);

	// shift out command until not busy is indicated
	bitvector in;
	do
	{
		in.clear();
		rc = jtag->shift(jtagAdapter::IRSHIFT, &shcmd, jtagAdapter::IDLE, &in);
        if (rc < 0)
            return -1;
		if ( (in.toNum() & 3) != 1)
			return -2;
	} while (in.toNum() & 4);

	lastCommand = cmd;
	return 0;
}

void MCU_AVR32::resetCPU(int onoff)
{
	// recover MCU from a bad state - ends with no reset, but halted
	if (onoff == 3)
	{
		resetCPU(1);
		_usleep(250*1000);
		haltCPU(1);
		resetCPU(0);
		_usleep(250*1000);
		return;
	}

	if (setCommand(0x0c) < 0) // AVR_RESET
        return;

	shiftData out;
	out.length = 5;
	out.tdi.push_back_value(onoff ? 0x1f : 0x00, 5);
	if (jtag->shift(jtagAdapter::DRSHIFT, &out, jtagAdapter::IDLE, NULL) < 0)
        return;
	jtag->clock(10);
}

void MCU_AVR32::haltCPU(int onoff)
{
	shiftData out;

	if (lastHaltState == onoff)
		return;

	if (setCommand(0x1c) < 0) // HALT
		return;

	out.length = 1;
	out.tdi.push_back_value(onoff, 1);
	if (jtag->shift(jtagAdapter::DRSHIFT, &out, jtagAdapter::IDLE, NULL) < 0)
        return;
	jtag->clock(10);
}

int MCU_AVR32::readAddrSingle (unsigned HSBaddr, enum XferSize opSize, uint32_t &data)
{
	if (setCommand(0x15) < 0) // MEMORY_SIZED_ACCESS
		return -4;

	shiftData out;
	out.length = 39;
	out.tdi.push_back_value(SAB_HSB,  4); //  4 - bus addr
	out.tdi.push_back_value(HSBaddr, 32); // 32 - address
	out.tdi.push_back_value(opSize,   2); //  2 - transfer size
	out.tdi.push_back_value(1,        1); //  1 - read

	bitvector in;
	int timeoutCount;
	// send address - repeat until busy is clear
	timeoutCount = 0;
	do
	{
		in.clear();
		int rc = jtag->shift(jtagAdapter::DRSHIFT, &out, jtagAdapter::DRUPDATE, &in);
        if (rc < 0)
        {
            printf("Jtag shift returned %d\n",rc);
            return -2;
        }
		if (timeoutCount++ > timeoutThreshold)
			return -3;
	} while (in[0]);

	// readback data - repeat until busy is clear
	out.tdi.clear();
	timeoutCount = 0;
	do
	{
		in.clear();
		int rc = jtag->shift(jtagAdapter::DRSHIFT, &out, jtagAdapter::DRUPDATE, &in);
        if (rc < 0)
        {
            printf("Jtag shift returned %d\n",rc);
            return -2;
        }
		if (in[33]) return -1; // error from addressphase
		if (timeoutCount++ > timeoutThreshold)
			return -3;
	} while (in[32]);

	// set data and return SUCCESS
	data = in.toNum();
	return 0;
}

int MCU_AVR32::writeAddrSingle(unsigned HSBaddr, enum XferSize opSize, uint32_t data)
{
	if (setCommand(0x15) < 0) // MEMORY_SIZED_ACCESS
		return -4;

	shiftData out;
	out.length = 39;
	out.tdi.push_back_value(SAB_HSB,  4); //  4 - bus addr
	out.tdi.push_back_value(HSBaddr, 32); // 32 - address
	out.tdi.push_back_value(opSize,   2); //  2 - transfer size
	out.tdi.push_back_value(0,        1); //  1 - write

	bitvector in;
	int timeoutCount;
	// send address - repeat until busy is clear
	timeoutCount = 0;
	do
	{
		in.clear();
		int rc = jtag->shift(jtagAdapter::DRSHIFT, &out, jtagAdapter::DRUPDATE, &in);
        if (rc < 0)
        {
            printf("Jtag shift returned %d\n",rc);
            return -2;
        }
		if (timeoutCount++ > timeoutThreshold)
			return -3;
	} while (in[0]);

	// send data - repeat until busy is clear
	out.tdi.clear();
	out.tdi.push_back_value(data, 32);
	out.tdi.push_back_value(0xff,  7);
	timeoutCount = 0;
	do
	{
		in.clear();
		int rc = jtag->shift(jtagAdapter::DRSHIFT, &out, jtagAdapter::DRUPDATE, &in);
        if (rc < 0)
        {
            printf("Jtag shift returned %d\n",rc);
            return -2;
        }
		if (in[1]) return -1; // error from addressphase
		if (timeoutCount++ > timeoutThreshold)
			return -3;
	} while (in[0]);

	out.tdi.clear();
	jtag->shift(jtagAdapter::DRSHIFT, &out, jtagAdapter::DRUPDATE, &in);
	if (in[1]) return -1; // error from dataphase

	return 0;
}

int MCU_AVR32::readAddrBlock(unsigned HSBaddr, uint32_t *data, unsigned count)
{
    int rc;
	rc = readAddrSingle(HSBaddr, Word, data[0]);
    if (rc < 0)
        return -1;

	if (setCommand(0x12) < 0) // MEMORY_BLOCK_ACCESS
		return -4;

	shiftData out;
	bitvector in;
	out.length = 34;

	for (unsigned i=1; i<count; i++)
	{
		int timeoutCount = 0;
		do
		{
			in.clear();
			rc = jtag->shift(jtagAdapter::DRSHIFT, &out, jtagAdapter::DRUPDATE, &in);
            if (rc < 0)
                return -1;
			if (in[33])
                return -1; // error in transfer
			if (timeoutCount++ > timeoutThreshold)
				return -3;
		} while (in[32]);
		//printf("@%08x: %s %08x\n", 4*i+HSBaddr, in.toString(bitvector::BINARY).c_str(), in.toNum() );
		data[i] = in.toNum();
	}

	return 0;
}

int MCU_AVR32::writeAddrBlock(unsigned HSBaddr, uint32_t *data, unsigned count)
{
    int rc;
	if (!count)
		return 1;

	if (writeAddrSingle(HSBaddr, Word, data[0]) < 0)
        return -1;

	if (setCommand(0x12) < 0) // MEMORY_BLOCK_ACCESS
		return -4;

	shiftData out;
	bitvector in;
	out.length = 34;

	for (unsigned i=1; i<count; i++)
	{
		out.tdi.clear();
		out.tdi.push_back_value(data[i], 32);
		out.tdi.push_back_value(0xff, 2);
		int timeoutCount = 0;

		do
		{
			in.clear();
			rc = jtag->shift(jtagAdapter::DRSHIFT, &out, jtagAdapter::DRUPDATE, &in);
            if (rc < 0)
                return -1;
			//printf("@%08x: %s %08x\n", 4*i+HSBaddr, in.toString(bitvector::BINARY).c_str(), in.toNum() );
            if (in[1])
                return -1; // error in transfer
			if (timeoutCount++ > timeoutThreshold)
				return -3;
		} while (in[0]);
	}

	return 0;
}

int MCU_AVR32::readNEXUS (uint8_t addr, uint32_t &data)
{
	if (setCommand(0x10) < 0) // NEXUS_ACCESS
		return -4;

	shiftData out;
	out.length = 34;
	out.tdi.push_back_value(addr,     7); //  7 - addr
	out.tdi.push_back_value(1,        1); //  1 - read
	out.tdi.push_back_value(-1U,     26); // 26 - dummy

	bitvector in;
	int timeoutCount;
	// send address - repeat until busy is clear
	timeoutCount = 0;
	do
	{
		in.clear();
		if (jtag->shift(jtagAdapter::DRSHIFT, &out, jtagAdapter::DRUPDATE, &in)<0)
			return -99;
		if (timeoutCount++ > timeoutThreshold)
			return -3;
	} while (in[0]);

	// readback data - repeat until busy is clear
	out.tdi.clear();
	timeoutCount = 0;
	do
	{
		in.clear();
		if (jtag->shift(jtagAdapter::DRSHIFT, &out, jtagAdapter::DRUPDATE, &in)<0)
			return -99;
		if (in[33]) return -1; // error from addressphase
		if (timeoutCount++ > timeoutThreshold)
			return -3;
	} while (in[32]);

	// set data and return SUCCESS
	data = in.toNum();
	return 0;
}

int MCU_AVR32::writeNEXUS (uint8_t addr, uint32_t data)
{
    int rc;
	if (setCommand(0x10) < 0) // NEXUS_ACCESS
		return -4;

	shiftData out;
	out.length = 34;
	out.tdi.push_back_value(addr,     7); //  7 - addr
	out.tdi.push_back_value(0,        1); //  0 - write
	out.tdi.push_back_value(-1U,     26); // 26 - dummy

	bitvector in;
	int timeoutCount;
	// send address - repeat until busy is clear
	timeoutCount = 0;
	do
	{
		in.clear();
		if (jtag->shift(jtagAdapter::DRSHIFT, &out, jtagAdapter::DRUPDATE, &in)<0)
			return -99;;
		if (timeoutCount++ > timeoutThreshold)
			return -3;
	} while (in[0]);

	// send data - repeat until busy is clear
	out.tdi.clear();
	out.tdi.push_back_value(data, 32);
	out.tdi.push_back_value(0xff,  2);
	timeoutCount = 0;
	do
	{
		in.clear();
		rc = jtag->shift(jtagAdapter::DRSHIFT, &out, jtagAdapter::DRUPDATE, &in);
        if (rc < 0)
            return -1;
		if (in[1])
            return -1; // error from addressphase
		if (timeoutCount++ > timeoutThreshold)
			return -3;
	} while (in[0]);

	out.tdi.clear();
	jtag->shift(jtagAdapter::DRSHIFT, &out, jtagAdapter::DRUPDATE, &in);
	if (in[1]) return -1; // error from dataphase

	return 0;
}

int MCU_AVR32::getFlashSize()
{
	uint32_t dat;
    int rc;

	if (isLvariant)
	{
		rc = readAddrSingle(FLASHCBase + MCU_AVR32::FPR_L, MCU_AVR32::Word, dat);
        if (rc < 0)
            return -1;
		switch (dat & 0x0f)
		{
			case  0: return    4 * 1024L;
			case  1: return    8 * 1024L;
			case  2: return   16 * 1024L;
			case  3: return   32 * 1024L;
			case  4: return   48 * 1024L;
			case  5: return   64 * 1024L;
			case  6: return   96 * 1024L;
			case  7: return  128 * 1024L;
			case  8: return  192 * 1024L;
			case  9: return  256 * 1024L;
			case 10: return  384 * 1024L;
			case 11: return  512 * 1024L;
			case 12: return  768 * 1024L;
			case 13: return 1024 * 1024L;
			case 14: return 2048 * 1024L;
		}
		return 0;
	}

	rc = readAddrSingle(FLASHCBase + MCU_AVR32::FSR, MCU_AVR32::Word, dat);
    if (rc < 0)
        return -1;
	dat >>= 13;
	dat  &= 0x7;

	switch (dat)
	{
		case 0: return   32 * 1024L;
		case 1: return   64 * 1024L;
		case 2: return  128 * 1024L;
		case 3: return  256 * 1024L;
		case 4: return  384 * 1024L;
		case 5: return  512 * 1024L;
		case 6: return  768 * 1024L;
		case 7: return 1024 * 1024L;
	}

	return 0;
}

int MCU_AVR32::getFrequency()
{
#if 0
#warning This function does not work, yet
	shiftData out;
	bitvector in;

	int div = 0xffff;

	setCommand(0x17); // SYNC_ACCESS

	out.length = 16;
	out.tdi.clear();
	out.tdi.push_back_value(div, 16);

	struct timeval tv_s, tv_e, tv_delta;
	gettimeofday(&tv_s, NULL);

	jtag->shift(jtagAdapter::DRSHIFT, &out, jtagAdapter::DRUPDATE, &in);
	out.tdi.clear();
	out.tdi.push_back_value(0, 16);

	do
	{
		in.clear();
		jtag->shift(jtagAdapter::DRSHIFT, &out, jtagAdapter::DRUPDATE, &in);
	} while (in.toNum() & 1);
	gettimeofday(&tv_e, NULL);

	timersub(&tv_e, &tv_s, &tv_delta);
	//printf("time: %u %u\n", tv_delta.tv_sec, tv_delta.tv_usec);

	return (1000000UL*div)/(tv_delta.tv_sec*1000000L + tv_delta.tv_usec);
#endif
	return -1;
}

void MCU_AVR32::flashWaitReady()
{
	uint32_t status;
	int timeoutCount = 0;
	do
	{
		if (readAddrSingle(FLASHCBase + MCU_AVR32::FSR, MCU_AVR32::Word, status) < 0)
            return;
		if (status & 0x3c)
			printf("AVR32_FSR = %x\n",status);
		if (timeoutCount++ > timeoutThreshold)
			return;
	}
	while (!(status & 1));
}

int MCU_AVR32::flashErasePage(int page)
{
	int hs = lastHaltState;

	haltCPU(1);

	// printf("eraseing page %d\n", page);
	if (writeAddrSingle(FLASHCBase + MCU_AVR32::FCMD, MCU_AVR32::Word, 0xa5000000 | (page<<8) | FLASH_EP) < 0)
        return -1;

	// we provoke a new setCommand to see if chain is still ok
    lastCommand = -1;
	flashWaitReady();

	haltCPU(hs);
	return 0;
}

int MCU_AVR32::flashWriteWords(uint32_t startAddr, uint32_t *data, unsigned count)
{
	int hs = lastHaltState;
	uint32_t pageOffset = (startAddr % pageSize) / 4;
	int page = (startAddr - 0x80000000) / pageSize;

	if (!(startAddr & 0x80000000))
		return -1;

	int isUP = startAddr & 0x00800000;

	haltCPU(1);

	while (count)
	{
		unsigned amount = (pageSize/4) - pageOffset;
		if (amount > count) amount = count;

		if (writeAddrSingle(FLASHCBase + MCU_AVR32::FCMD, MCU_AVR32::Word, 0xa5000000 | (page<<8) | FLASH_CPB) < 0)
			return -2;
		flashWaitReady();

		//printf("write: %x %d %d\n", startAddr, page, amount);
		if (writeAddrBlock(startAddr, data, amount) < 0)
			return -3;

		if (writeAddrSingle(FLASHCBase + MCU_AVR32::FCMD, MCU_AVR32::Word, 0xa5000000 | (page<<8) | (isUP ? FLASH_WUP : FLASH_WP)) < 0)
			return -4;

		// we provoke a new setCommand to see if chain is still ok
		lastCommand = -1;
		flashWaitReady();

		startAddr += 4*amount;
		data      += amount;
		count     -= amount;
		page++;
		pageOffset = 0;
	}

	haltCPU(hs);

	return 0;
}

MCU_AVR32::MCU_AVR32(jtagAdapter *adapter)
{
	jtag = adapter;
	lastCommand = -1;
	lastHaltState = -1;
	timeoutThreshold = 1000;
}

MCU_AVR32::~MCU_AVR32()
{
	jtag->doState(jtagAdapter::RESET);
	haltCPU(0);
}

int MCU_AVR32::init(int reset, int unlock)
{
	if (jtag->init() < 0)
		return -1;

	jtag->setJTAGSpeed(20000000); // 20 MHz
	jtag->doState(jtagAdapter::RESET);

	// recover from unclean states
	if (reset)
		resetCPU(3);

	uint32_t id;
	readNEXUS(MCU_AVR32::NEXUS_DID, id);

	if (unlock)
	{
		// AT32UC3Lxx behaves differently
		if ((id & 0x0200003f) == 0x0200003f)
		{
			isLvariant = 1;
			FLASHCBase = PBB + 0x0;
			pageSize = 0x100;

			// check and clear BOOTPROT
			uint32_t fuse_low;
			readAddrSingle(FLASHCBase + MCU_AVR32::FGPFRLO_L, MCU_AVR32::Word, fuse_low);
			printf("FUSE LOW = %x\n",fuse_low);
			if (((fuse_low >> 17) & 0x07) != 0x07)
			{
				printf("Disabling active bootloader protection\n");
				writeAddrSingle(FLASHCBase + MCU_AVR32::FCMD, MCU_AVR32::Word, 0xa5000000 | FLASH_EGPB | (17<<8));
				writeAddrSingle(FLASHCBase + MCU_AVR32::FCMD, MCU_AVR32::Word, 0xa5000000 | FLASH_EGPB | (18<<8));
				writeAddrSingle(FLASHCBase + MCU_AVR32::FCMD, MCU_AVR32::Word, 0xa5000000 | FLASH_EGPB | (19<<8));
			}
		}
		else
		{
			isLvariant = 0;
			FLASHCBase = PBB + 0x1400;
			pageSize = 0x200;

			// check and clear BOOTPROT
			uint32_t fuse_low;
			readAddrSingle(FLASHCBase + MCU_AVR32::FGPFRLO_AB, MCU_AVR32::Word, fuse_low);
			if (((fuse_low >> 17) & 0x07) != 0x07)
			{
				printf("Disabling active bootloader protection\n");
				writeAddrSingle(FLASHCBase + MCU_AVR32::FCMD, MCU_AVR32::Word, 0xa5000000 | FLASH_EGPB | (17<<8));
				writeAddrSingle(FLASHCBase + MCU_AVR32::FCMD, MCU_AVR32::Word, 0xa5000000 | FLASH_EGPB | (18<<8));
				writeAddrSingle(FLASHCBase + MCU_AVR32::FCMD, MCU_AVR32::Word, 0xa5000000 | FLASH_EGPB | (19<<8));
			}
		}
	} // unlock

	return 0;
}

int  MCU_AVR32::flashRead(unsigned addr, unsigned bytes, uint8_t *data)
{
	unsigned words = (bytes+3)/4;
	uint32_t *readb = new uint32_t[words];

	int ret = readAddrBlock(addr, readb, words);
	for (unsigned i=0; i<words; i++)
		readb[i] = htonl(readb[i]);

	memcpy(data, (uint8_t*)readb, bytes);
	delete[] readb;

	return ret;
}

int MCU_AVR32::flashWrite(unsigned addr, unsigned bytes, uint8_t *data)
{
	unsigned words = (bytes+3)/4;
	uint32_t *writedata = new uint32_t[words];

	memcpy((uint8_t*)writedata, data, bytes);

	for (unsigned i=0; i<words; i++)
		writedata[i] = htonl(writedata[i]);

	int ret = flashWriteWords(addr, writedata, words);

	delete[] writedata;

	return ret;
}

int  MCU_AVR32::flashErase(unsigned addr)
{
	int ret = flashErasePage( (addr - 0x80000000)/getPageSize() );
	return ret;
}

int  MCU_AVR32::flashChipErase()
{
	if (setCommand(0x0f) < 0) // CHIP_ERASE
		return -1;

	return 0;
}

