/** -*- c++ -*- **************************************************************
 **
 ** Copyright 2014 Daniel Paschka (daniel@drachenschuppe.de)
 **
 ** Licensed under the Apache License, Version 2.0 (the "License");
 ** you may not use this file except in compliance with the License.
 ** You may obtain a copy of the License at
 **
 ** http://www.apache.org/licenses/LICENSE-2.0
 **
 ** Unless required by applicable law or agreed to in writing, software
 ** distributed under the License is distributed on an "AS IS" BASIS,
 ** WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 ** See the License for the specific language governing permissions and
 ** limitations under the License.
 *****************************************************************************/

#ifndef MCU_AVR32_H
#define MCU_AVR32_H

#include "jtagAdapter.h"

/*
 *  opSize  : 0 - word, 1 - hword, 2 - byte
 *  HSBaddr : address on the High Speed Bus
 */

class MCU_AVR32
{
protected:
	int timeoutThreshold;
	int lastCommand;
	int lastHaltState;
	int pageSize;
	int isLvariant;
	uint32_t FLASHCBase;
	int setCommand(int cmd);

	jtagAdapter *jtag;

public:
	enum
	{
		time_sectorerase =  8,
		time_sectorwrite = 64,
		time_sectorread  = 25,
	};

	enum HSB_Addr
	{
		Ram       = 0x00000000,
		FlashData = 0x80000000,
		PBB       = 0xFFFE0000,
	};

	enum FLASHC_Offset
	{
		FCR        = 0x00,
		FCMD       = 0x04,
		FSR        = 0x08,
		FGPFRHI_AB = 0x0c,
		FGPFRLO_AB = 0x10,

		FPR_L      = 0x0c,
		FVR_L      = 0x10,
		FGPFRHI_L  = 0x14,
		FGPFRLO_L  = 0x18,
	};

	enum FLASH_Cmd
	{
		FLASH_NOP   =  0,  // No Operation
		FLASH_WP    =  1,  // Write Page
		FLASH_EP    =  2,  // Erase Page
		FLASH_CPB   =  3,  // Clear Page Buffer
		FLASH_EA    =  6,  // Erase All
		FLASH_EGPB  =  8,  // Erase General Purpose Fuse Bit
		FLASH_QPR   = 12,  // Quick Page Read
		FLASH_WUP   = 13,  // Write User Page
		FLASH_EUP   = 14,  // Erase User Page
		FLASH_QPRUP = 15   // Quick Page Read User Page
	};

	enum XferSize
	{
		Byte     = 0,
		HalfWord = 1,
		Word     = 2,
	};

	enum NEXUS_Addr
	{
		NEXUS_DID       =  0,
		NEXUS_DC        =  2,  // p. 105
		NEXUS_DS        =  4,  // p. 108
		NEXUS_RWCS      =  7,
		NEXUS_RWA       =  9,
		NEXUS_RWD       = 10,
		NEXUS_WT        = 11,
		NEXUS_DTC       = 13,
		NEXUS_DTSA0     = 14,
		NEXUS_DTSA1     = 15,
		NEXUS_DTEA0     = 18,
		NEXUS_DTEA1     = 19,

		// Breakpoint/Watchpoint Ctrl
		NEXUS_BWC0A     = 22,
		NEXUS_BWC0B     = 23,
		NEXUS_BWC1A     = 24,
		NEXUS_BWC1B     = 25,
		NEXUS_BWC2A     = 26,
		NEXUS_BWC2B     = 27,
		NEXUS_BWC3A     = 28,
		NEXUS_BWC3B     = 29,
		// PC Breakpoint/Watchpoint Address
		NEXUS_BWA0A     = 30,
		NEXUS_BWA0B     = 31,
		NEXUS_BWA1A     = 32,
		NEXUS_BWA1B     = 33,
		NEXUS_BWA2A     = 34,
		NEXUS_BWA2B     = 35,
		// Data Breakpoint/Watchpoint Address
		NEXUS_BWA3A     = 36,
		NEXUS_BWA3B     = 37,
		// Breakpoint/Watchpoint Data
		NEXUS_BWD3A     = 38,
		NEXUS_BWD3B     = 39,

		NEXUS_NXCFG     = 64,   // p. 103
		NEXUS_DINST     = 65,   // p. 109
		NEXUS_DPC       = 66,   // p. 109
		NEXUS_CPUCtrlMask = 67,
		NEXUS_DCCPU     = 68,   // p. 104
		NEXUS_DCEMU     = 69,   // p. 104
		NEXUS_DCSR      = 70,   // p. 105
		NEXUS_PID       = 71,   // p. 141
	};

	// p.105
	enum
	{
		NEXUS_DC_RES    = (1<<30),
		NEXUS_DC_IFM    = (1<<24),
		NEXUS_DC_IRP    = (1<<23),
		NEXUS_DC_DBE    = (1<<13),
		NEXUS_DC_DBR    = (1<<12),
		NEXUS_DC_SS     = (1<< 8),
	};

	// p.108
	enum
	{
		NEXUS_DS_DBA    = (1<<26),
		NEXUS_DS_BOZ    = (1<<25),
		NEXUS_DS_INC    = (1<<24),
		NEXUS_DS_BP3B   = (1<<15),
		NEXUS_DS_BP3A   = (1<<14),
		NEXUS_DS_BP2B   = (1<<13),
		NEXUS_DS_BP2A   = (1<<12),
		NEXUS_DS_BP1B   = (1<<11),
		NEXUS_DS_BP1A   = (1<<10),
		NEXUS_DS_BP0B   = (1<< 9),
		NEXUS_DS_BP0A   = (1<< 8),
		NEXUS_DS_DBS    = (1<< 5),
		NEXUS_DS_HWE    = (1<< 3),
		NEXUS_DS_HWB    = (1<< 2),
		NEXUS_DS_SWB    = (1<< 1),
		NEXUS_DS_SSS    = (1<< 0),
	};

	int  getFrequency();
	void resetCPU(int onoff);
	void haltCPU(int onoff);

	int  readAddrSingle (unsigned HSBaddr, enum XferSize opSize, uint32_t &data);
	int  writeAddrSingle(unsigned HSBaddr, enum XferSize opSize, uint32_t  data);
	int  readAddrBlock  (unsigned HSBaddr, uint32_t *data, unsigned count); // count is in words
	int  writeAddrBlock (unsigned HSBaddr, uint32_t *data, unsigned count); // count is in words
	int  readNEXUS  (uint8_t addr, uint32_t &data);
	int  writeNEXUS (uint8_t addr, uint32_t data);

	int  getFlashSize();
	void flashWaitReady();
	int  flashErasePage(int page);
	int  flashWriteWords(uint32_t startAddr, uint32_t *data, unsigned count);

	MCU_AVR32(jtagAdapter *adapter);
	~MCU_AVR32();

	int  init(int reset, int unlock);

	unsigned getPageSize () { return pageSize; }
	int  flashRead (unsigned addr, unsigned bytes, uint8_t *data);
	int  flashWrite(unsigned addr, unsigned bytes, uint8_t *data);
	int  flashErase(unsigned addr);
	int  flashChipErase();
};

#endif // _MCU_AVR32_H

